"""Decryption module for Theengs Gateway.

This module adds decryptors for encrypted advertisements of a selection of
devices.
"""
from __future__ import annotations

import binascii
from abc import ABC, abstractmethod

from Cryptodome.Cipher import AES


class AdvertisementDecryptor(ABC):
    """Abstract class that represents a decryptor for advertisements."""

    @abstractmethod
    def compute_nonce(self, address: str, decoded_json: dict) -> bytes:
        """Compute the nonce for a specific address and JSON input."""

    @abstractmethod
    def decrypt(
        self,
        bindkey: bytes,
        address: str,
        decoded_json: dict,
    ) -> bytes:
        """Decrypt ciphertext from JSON input."""

    @abstractmethod
    def replace_encrypted_data(
        self,
        decrypted_data: bytes,
        data_json: dict,
        decoded_json: dict,
    ) -> None:
        """Replace the encrypted data by decrypted payload."""


class LYWSD03MMC_PVVXDecryptor(AdvertisementDecryptor):  # noqa: N801
    """Class for decryption of LYWSD03MMX PVVX encrypted advertisements."""

    def compute_nonce(self, address: str, decoded_json: dict) -> bytes:
        """Compute the nonce for a specific address and JSON input."""
        # The nonce consists of:
        # 6 bytes: device address in reverse
        # 1 byte : length of (service data + type and UUID)
        # 1 byte : "16" -> AD type for "Service Data - 16-bit UUID"
        # 2 bytes: "1a18" -> UUID 181a in little-endian
        # 1 byte : counter
        return binascii.unhexlify(
            "".join(
                [
                    reverse_address(address),
                    bytes([len(decoded_json["servicedata"]) // 2 + 3]).hex(),
                    "161a18",
                    decoded_json["ctr"],
                ],
            ),
        )

    def decrypt(
        self,
        bindkey: bytes,
        address: str,
        decoded_json: dict,
    ) -> bytes:
        """Decrypt ciphertext from JSON input with AES CCM."""
        nonce = self.compute_nonce(address, decoded_json)
        cipher = AES.new(bindkey, AES.MODE_CCM, nonce=nonce, mac_len=4)
        cipher.update(b"\x11")
        payload = bytes.fromhex(decoded_json["cipher"])
        mic = bytes.fromhex(decoded_json["mic"])
        return cipher.decrypt_and_verify(payload, mic)

    def replace_encrypted_data(
        self,
        decrypted_data: bytes,
        data_json: dict,
        decoded_json: dict,  # noqa: ARG002
    ) -> None:
        """Replace the encrypted data by decrypted payload."""
        data_json["servicedata"] = decrypted_data.hex()


class BTHomeV2Decryptor(AdvertisementDecryptor):
    """Class for decryption of BTHome v2 encrypted advertisements."""

    def compute_nonce(self, address: str, decoded_json: dict) -> bytes:
        """Compute the nonce for a specific address and JSON input."""
        # The nonce consists of:
        # 6 bytes: device address
        # 2 bytes: "d2fc" -> UUID fcd2 in little-endian
        # 1 byte : BTHome device info
        # 1 byte : counter
        return binascii.unhexlify(
            "".join(
                [
                    address.replace(":", ""),
                    "d2fc",  # UUID fcd2
                    decoded_json["servicedata"][:2],
                    decoded_json["ctr"],
                ],
            ),
        )

    def decrypt(
        self,
        bindkey: bytes,
        address: str,
        decoded_json: dict,
    ) -> bytes:
        """Decrypt ciphertext from JSON input with AES CCM."""
        nonce = self.compute_nonce(address, decoded_json)
        cipher = AES.new(bindkey, AES.MODE_CCM, nonce=nonce, mac_len=4)
        payload = bytes.fromhex(decoded_json["cipher"])
        mic = bytes.fromhex(decoded_json["mic"])
        return cipher.decrypt_and_verify(payload, mic)

    def replace_encrypted_data(
        self,
        decrypted_data: bytes,
        data_json: dict,
        decoded_json: dict,
    ) -> None:
        """Replace the encrypted data by decrypted payload."""
        # Clear encryption and MAC included bits in device info
        # See https://bthome.io/format/
        device_info = bytes.fromhex(decoded_json["servicedata"][:2])
        mask = 0b11111100
        masked_device_info = int.from_bytes(device_info, "big") & mask
        bthome_service_data = bytearray(masked_device_info.to_bytes(1, "big"))

        # Replace encrypted data by decrypted payload
        bthome_service_data.extend(decrypted_data)
        data_json["servicedata"] = bthome_service_data.hex()


_DECRYPTORS = {
    1: LYWSD03MMC_PVVXDecryptor,
    2: BTHomeV2Decryptor,
}


class UnsupportedEncryptionError(Exception):
    """Exception raised when trying to decrypt an unsupported device."""


def create_decryptor(mode: int) -> AdvertisementDecryptor:
    """Return the decryptor class for the given encryption mode."""
    if mode not in _DECRYPTORS:
        raise UnsupportedEncryptionError
    return _DECRYPTORS[mode]()  # type: ignore[abstract]


def reverse_address(address: str) -> str:
    """Reverse device address bytes and remove colons."""
    return "".join(address.split(":")[::-1])



"""TheengsGateway - Decode things and devices and publish data to an MQTT broker.

Copyright: (c)Florian ROBERT

This file is part of TheengsGateway.

TheengsGateway is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TheengsGateway is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# mypy: disable-error-code=attr-defined
from __future__ import annotations

import json
import re
from time import time

from TheengsDecoder import getProperties

from .ble_gateway import DataJSONType, Gateway, logger

ha_dev_classes = [
    "battery",
    "carbon_dioxide",
    "carbon_monoxide",
    "current",
    "data_size",
    "distance",
    "door",
    "duration",
    "energy",
    "enum",
    "gas",
    "humidity",
    "illuminance",
    "irradiance",
    "lock",
    "motion",
    "moving",
    "pm10",
    "pm25",
    "power",
    "power_factor",
    "pressure",
    "problem",
    "restart",
    "signal_strength",
    "temperature",
    "timestamp",
    "voltage",
    "water",
    "weight",
    "window",
]

ha_dev_units = [
    "W",
    "kW",
    "V",
    "kWh",
    "A",
    "W",
    "°C",
    "°F",
    "ms",
    "s",
    "min",
    "hPa",
    "L",
    "kg",
    "lb",
    "µS/cm",
    "ppm",
    "μg/m³",
    "m³",
    "mg/m³",
    "m/s²",
    "lx",
    "Ω",
    "%",
    "bar",
    "bpm",
    "dB",
    "dBm",
    "B",
    "UV index",
    "m/s",
    "km/h",
    "°",
    "mm",
    "mm/h",
    "cm",
]


class DiscoveryGateway(Gateway):
    """BLE to MQTT gateway class with Home Assistant MQTT discovery."""

    def __init__(self, configuration: dict) -> None:
        super().__init__(configuration)
        self.discovered_entities: list[str] = []

    def publish_device_info(self, pub_device) -> None:  # noqa: ANN001
        """Publish sensor directly to Home Assistant via MQTT discovery."""
        pub_device_uuid = pub_device["id"].replace(":", "")

        pub_device_copy = self.copy_pub_device(pub_device)

        device_data = json.dumps(pub_device_copy)
        if (
            pub_device_uuid in self.discovered_entities
            or pub_device["model_id"] in self.configuration["discovery_filter"]
        ):
            logger.debug("Already discovered or filtered: %s", pub_device_uuid)
            self.publish(
                device_data,
                self.configuration["publish_topic"] + "/" + pub_device_uuid,
            )
            if self.configuration["presence"]:
                self.publish(device_data, self.configuration["presence_topic"])
            return

        logger.info("publishing device `%s`", pub_device)
        pub_device["properties"] = json.loads(
            getProperties(pub_device["model_id"]),
        )["properties"]

        hadevice = self.prepare_hadevice(pub_device_uuid, pub_device)

        discovery_topic = self.configuration["discovery_topic"]
        state_topic = self.build_state_topic(pub_device)

        data = getProperties(pub_device["model_id"])
        data = json.loads(data)
        data = data["properties"]
        entity_type = "sensor"

        for k in data:
            device = {}
            device["stat_t"] = state_topic
            # device_tracker discovery
            self.publish_device_tracker(
                pub_device_uuid,
                state_topic,
                pub_device,
                hadevice,
            )
            # If the properties key is "mac" or "device", skip its discovery
            if k in {"mac", "device"}:
                continue
            if k in pub_device["properties"]:
                if pub_device["properties"][k]["name"] in ha_dev_classes:
                    device["dev_cla"] = pub_device["properties"][k]["name"]
                if pub_device["properties"][k]["unit"] in ha_dev_units:
                    device["unit_of_meas"] = pub_device["properties"][k]["unit"]
                    device["state_class"] = "measurement"
                    entity_type = "sensor"
                elif pub_device["properties"][k]["unit"] == "status":
                    entity_type = "binary_sensor"
                    device["pl_on"] = "True"
                    device["pl_off"] = "False"
            device["name"] = pub_device["model_id"] + "-" + k
            device["uniq_id"] = pub_device_uuid + "-" + k
            if k == "unlocked":
                device[
                    "val_tpl"
                ] = "{% if value_json.get('unlocked') is true -%}True{%- else -%}False{%- endif %}"  # noqa: E501
            elif self.configuration["hass_discovery"]:
                device["val_tpl"] = "{{ value_json." + k + " | is_defined }}"
            else:
                device["val_tpl"] = "{{ value_json." + k + " }}"

            config_topic = (
                discovery_topic
                + "/"
                + entity_type
                + "/"
                + pub_device_uuid
                + "-"
                + k
                + "/config"
            )
            device["device"] = hadevice  # type: ignore[assignment]
            self.publish(json.dumps(device), config_topic, retain=True)

        self.discovered_entities.append(pub_device_uuid)
        self.publish_device_data(
            pub_device_uuid,
            device_data,
        )

    def publish_device_data(self, uuid: str, data: str) -> None:
        """Publish device data to the configured MQTT topics."""
        self.publish(data, f"{self.configuration['publish_topic']}/{uuid}")
        if self.configuration["presence"]:
            self.publish(data, self.configuration["presence_topic"])

    def prepare_hadevice(self, uuid: str, device: dict) -> dict:
        """Prepare Home Assistant device configuration."""
        return {
            "ids": [uuid],
            "cns": [["mac", uuid]],
            "mf": device["brand"],
            "mdl": device["model_id"],
            "name": device["model"] + "-" + uuid[6:],
            "via_device": self.configuration.get(
                "discovery_device_name",
                "Unknown",
            ),
        }

    def build_state_topic(self, device: dict) -> str:
        """Build state topic."""
        state_topic = (
            self.configuration["publish_topic"] + "/" + device["id"].replace(":", "")
        )
        return re.sub(
            r".+?/",
            "+/",
            state_topic,
            count=len(re.findall(r"/", state_topic)) - 1,
        )

    def publish_device_tracker(
        self,
        pub_device_uuid: str,
        state_topic: str,
        pub_device: DataJSONType,
        hadevice: dict,
    ) -> None:
        """Publish device_tracker discovery."""
        if "track" in pub_device:
            config_topic = (
                self.configuration["discovery_topic"]
                + "/device_tracker/"
                + pub_device_uuid
                + "-tracker"
                + "/config"
            )

            tracker = {}
            tracker["stat_t"] = state_topic
            tracker["name"] = pub_device["model_id"] + "-tracker"  # type: ignore[assignment,operator]
            tracker["uniq_id"] = pub_device_uuid + "-tracker"
            tracker[
                "val_tpl"
            ] = "{% if value_json.get('rssi') -%}home{%- else -%}not_home{%- endif %}"
            tracker["source_type"] = "bluetooth_le"
            tracker["device"] = hadevice  # type: ignore[assignment]

            self.publish(json.dumps(tracker), config_topic, retain=True)

    def copy_pub_device(self, device: dict) -> dict:
        """Copy pub_device and remove "track" if publish_advdata is false."""
        # Update tracker last received time
        if "track" in device:
            self.discovered_trackers[device["id"]] = round(time())
            logger.debug("Discovered Trackers: %s", self.discovered_trackers)

        pub_device_copy = device.copy()
        # Remove "track" if PUBLISH_ADVDATA is 0
        if not self.configuration["publish_advdata"] and "track" in pub_device_copy:
            pub_device_copy.pop("track", None)

        return pub_device_copy



"""Diagnose module for Theengs Gateway.

This module can be run on the command line with python -m TheengsGateway.diagnose
to show diagnostic information for debugging purposes.
"""
from __future__ import annotations

import argparse
import asyncio
import json
import platform
import re
import sys
from pathlib import Path
from typing import Dict, List, Union

from importlib_metadata import PackageNotFoundError, version

ConfigType = Dict[str, Union[str, int, List[str]]]
_ADDR_RE = re.compile(r"^(([0-9A-F]{2}:){3})([0-9A-F]{2}:){2}[0-9A-F]{2}$")


def _anonymize_strings(fields: list[str], config: ConfigType) -> None:
    for field in fields:
        if field in config:
            config[field] = "***"


def _anonymize_address(address: str) -> str:
    addr_parts = _ADDR_RE.match(address)
    if addr_parts:
        return f"{addr_parts.group(1)}XX:XX:XX"
    return "INVALID ADDRESS"


def _anonymize_addresses(addresses: list[str]) -> list[str]:
    return [_anonymize_address(address) for address in addresses]


def _anonymize_addr_keys(addr_keys: dict[str, str]) -> dict[str, str]:
    """Anonymize the addresses and corresponding keys in a dictionary."""
    return {_anonymize_address(address): "***" for address in addr_keys}


# This function is taken from Textual
def _section(title: str, values: dict[str, str]) -> None:
    """Print a collection of named values within a titled section."""
    max_name = max(map(len, values.keys()))
    max_value = max(map(len, [str(value) for value in values.values()]))
    print(f"## {title}")
    print()
    print(f"| {'Name':{max_name}} | {'Value':{max_value}} |")
    print(f"|-{'-' * max_name}-|-{'-'*max_value}-|")
    for name, value in values.items():
        print(f"| {name:{max_name}} | {value!s:{max_value}} |")
    print()


def _versions() -> None:
    """Print useful version numbers."""
    try:
        packages = {
            "Theengs Gateway": version("TheengsGateway"),
            "Theengs Decoder": version("TheengsDecoder"),
            "Bleak": version("bleak"),
            "Bluetooth Clocks": version("bluetooth-clocks"),
            "Bluetooth Numbers": version("bluetooth-numbers"),
            "Paho MQTT": version("paho-mqtt"),
        }
    except PackageNotFoundError as e:
        print(f"Package {e.name} not found. Please install it with:")
        print()
        print(f"    pip install {e.name}")
        print()

    if sys.version_info[:2] >= (3, 9):
        try:
            packages["Bluetooth Adapters"] = version("bluetooth-adapters")
        except PackageNotFoundError as e:
            print(f"Package {e.name} not found. Please install it with:")
            print()
            print(f"    pip install {e.name}")
            print()

    _section("Package Versions", packages)


def _python() -> None:
    """Print information about Python."""
    _section(
        "Python",
        {
            "Version": platform.python_version(),
            "Implementation": platform.python_implementation(),
            "Compiler": platform.python_compiler(),
            "Executable": sys.executable,
        },
    )


def _os() -> None:
    """Print operating system information."""
    os_parameters = {
        "System": platform.system(),
        "Release": platform.release(),
        "Version": platform.version(),
        "Machine type": platform.machine(),
    }
    if platform.system() == "Linux" and sys.version_info[:2] >= (3, 10):
        os_release = platform.freedesktop_os_release()  # type: ignore[attr-defined]
        os_parameters["Distribution"] = os_release["PRETTY_NAME"]
    _section("Operating System", os_parameters)


def _config(config_path: Path) -> None:
    """Print the anonymized Theengs Gateway configuration."""
    print("## Configuration")
    print()
    print(f"File: {config_path.resolve()}")
    print()
    try:
        with config_path.open(encoding="utf-8") as config_file:
            config = json.load(config_file)
            _anonymize_strings(["user", "pass"], config)
            config["time_sync"] = _anonymize_addresses(config["time_sync"])
            config["bindkeys"] = _anonymize_addr_keys(config["bindkeys"])
            config["identities"] = _anonymize_addr_keys(config["identities"])
        print("```")
        print(json.dumps(config, sort_keys=True, indent=4))
        print("```")
        print()
    except FileNotFoundError:
        print("Configuration file not found")
        print()
    except json.JSONDecodeError as exception:
        print(f"Malformed JSON configuration file: {exception}")
        print()


async def _adapters() -> None:
    """Print information about the system's Bluetooth adapters."""
    if sys.version_info[:2] >= (3, 9):
        from bluetooth_adapters import get_adapters

        print("## Bluetooth adapters")
        print()
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        print(f"Default adapter: {bluetooth_adapters.default_adapter}")
        print()

        for adapter, properties in sorted(bluetooth_adapters.adapters.items()):
            properties["address"] = _anonymize_address(properties["address"])
            print("#", end="")
            _section(adapter, properties)  # type: ignore[arg-type]


async def diagnostics(config_path: Path) -> None:
    """Main function of the diagnose module.

    This function prints a header and various sections with diagnostic information
    about package versions, Python and operating system information, the
    anonymized configuration file and Bluetooth adapter information in Markdown
    format.
    """
    print("# Theengs Gateway Diagnostics")
    print()
    _versions()
    _python()
    _os()
    _config(config_path)
    await _adapters()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="Path to the configuration file (default: ~/theengsgw.conf)",
    )
    args = parser.parse_args()

    if args.config:
        config_path = Path(args.config)
    else:
        config_path = Path("~/theengsgw.conf").expanduser()

    asyncio.run(diagnostics(config_path))



"""TheengsGateway - Decode things and devices and publish data to an MQTT broker.

Copyright: (c)Florian ROBERT

This file is part of TheengsGateway.

TheengsGateway is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TheengsGateway is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
from pathlib import Path

from .ble_gateway import run
from .config import (
    DEFAULT_CONFIG,
    merge_args_with_config,
    parse_args,
    read_configuration,
    write_configuration,
)


def main() -> None:
    """Main entry point of the TheengsGateway program."""
    args = parse_args()
    config_path = Path(args.config)

    configuration = read_configuration(config_path)
    # Merge default configuration, with data read from the configuration file
    # overriding default data.
    # This guarantees that all keys we refer to are in the dictionary.
    configuration = {**DEFAULT_CONFIG, **configuration}
    merge_args_with_config(configuration, args)

    # Remove /sensor if existing in the configuration as we now handle different types
    # of devices.
    if configuration["discovery_topic"].endswith("/sensor"):
        configuration["discovery_topic"] = configuration["discovery_topic"][:-7]

    if not configuration["host"]:
        sys.exit("MQTT host is not specified")

    write_configuration(configuration, config_path)
    run(configuration, config_path)



"""Configuration module for Theengs Gateway.

This module handles everything that has to do with configuration files and
command-line parameters for Theengs Gateway.
"""
from __future__ import annotations

import argparse
import json
from pathlib import Path

from importlib_metadata import version

# Each configuration option is added to:
# - the DEFAULT_CONFIG dict with its default value
# - the parse_args function for its command-line argument

DEFAULT_CONFIG = {
    "host": "",
    "port": 1883,
    "user": "",
    "pass": "",
    "ble_scan_time": 5,
    "ble_time_between_scans": 5,
    "publish_topic": "home/TheengsGateway/BTtoMQTT",
    "lwt_topic": "home/TheengsGateway/LWT",
    "subscribe_topic": "home/+/BTtoMQTT/undecoded",
    "presence_topic": "home/TheengsGateway/presence",
    "presence": 0,
    "publish_all": 1,
    "log_level": "INFO",
    "discovery": 1,
    "hass_discovery": 1,
    "general_presence": 0,
    "discovery_topic": "homeassistant",
    "discovery_device_name": "TheengsGateway",
    "discovery_filter": [
        "IBEACON",
    ],
    "adapter": "",
    "scanning_mode": "active",
    "time_sync": [],
    "time_format": 0,
    "publish_advdata": 0,
    "bindkeys": {},
    "enable_tls": 0,
    "enable_websocket": 0,
    "identities": {},
    "tracker_timeout": 120,
    "ble": 1,
}


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments and return them."""
    parser = argparse.ArgumentParser(
        description=f"TheengsGateway {version('TheengsGateway')}",
    )
    parser.add_argument(
        "-a",
        "--adapter",
        type=str,
        help="Bluetooth adapter (e.g. hci1 on Linux)",
    )
    parser.add_argument(
        "-b",
        "--ble",
        type=int,
        help="Enable (1) or disable (0) BLE (default: 1)",
    )
    parser.add_argument(
        "-bk",
        "--bindkeys",
        nargs="+",
        metavar=("ADDRESS", "BINDKEY"),
        help="Device addresses and their bindkeys: ADDR1 KEY1 ADDR2 KEY2",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=str(Path("~/theengsgw.conf").expanduser()),
        help="Path to the configuration file (default: ~/theengsgw.conf)",
    )
    parser.add_argument(
        "-D",
        "--discovery",
        type=int,
        help="Enable(1) or disable(0) MQTT discovery",
    )
    parser.add_argument(
        "-Df",
        "--discovery_filter",
        nargs="+",
        help="Device discovery filter list for Home Assistant",
    )
    parser.add_argument(
        "-Dh",
        "--hass_discovery",
        type=int,
        help="Enable(1) or disable(0) Home Assistant MQTT discovery (default: 1)",
    )
    parser.add_argument(
        "-Dn",
        "--discovery_name",
        type=str,
        help="Device name for Home Assistant",
    )
    parser.add_argument(
        "-Dt",
        "--discovery-topic",
        type=str,
        help="MQTT Discovery topic",
    )
    parser.add_argument(
        "-Gp",
        "--general_presence",
        type=int,
        help="Enable (1) or disable (0) general present/absent presence when --discovery: 0 (default: 0)",  # noqa: E501
    )
    parser.add_argument(
        "-H",
        "--host",
        type=str,
        help="MQTT host address",
    )
    parser.add_argument(
        "-id",
        "--identities",
        nargs="+",
        metavar=("ADDRESS", "IRK"),
        help="Identity addresses and their IRKs: ADDR1 IRK1 ADDR2 IRK2",
    )
    parser.add_argument(
        "-Lt",
        "--lwt_topic",
        type=str,
        help="MQTT LWT topic",
    )
    parser.add_argument(
        "-ll",
        "--log_level",
        type=str,
        help="TheengsGateway log level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "-P",
        "--port",
        type=int,
        help="MQTT host port",
    )
    parser.add_argument(
        "-p",
        "--pass",
        type=str,
        help="MQTT password",
    )
    parser.add_argument(
        "-pa",
        "--publish_all",
        type=int,
        help="Publish all (1) or only decoded (0) advertisements (default: 1)",
    )
    parser.add_argument(
        "-padv",
        "--publish_advdata",
        type=int,
        help="Publish advertising and advanced data (1) or not (0) (default: 0)",
    )
    parser.add_argument(
        "-pr",
        "--presence",
        type=int,
        help="Enable (1) or disable (0) presence publication (default: 1)",
    )
    parser.add_argument(
        "-prt",
        "--presence_topic",
        type=str,
        help="MQTT presence topic",
    )
    parser.add_argument(
        "-pt",
        "--pub_topic",
        type=str,
        help="MQTT publish topic",
    )
    parser.add_argument(
        "-s",
        "--scanning_mode",
        type=str,
        choices=("active", "passive"),
        help="Scanning mode (default: active)",
    )
    parser.add_argument(
        "-sd",
        "--scan_duration",
        type=int,
        help="BLE scan duration (seconds)",
    )
    parser.add_argument(
        "-st",
        "--sub_topic",
        type=str,
        help="MQTT subscribe topic",
    )
    parser.add_argument(
        "-tb",
        "--time_between",
        type=int,
        help="Seconds to wait between scans",
    )
    parser.add_argument(
        "-tf",
        "--time_format",
        type=int,
        help="Use 12-hour (1) or 24-hour (0) time format for clocks (default: 0)",
    )
    parser.add_argument(
        "-tls",
        "--enable_tls",
        type=int,
        help="Enable (1) or disable (0) TLS (default: 0)",
    )
    parser.add_argument(
        "-to",
        "--tracker_timeout",
        type=int,
        help="Tracker timeout duration (seconds)",
    )
    parser.add_argument(
        "-ts",
        "--time_sync",
        nargs="+",
        help="Addresses of Bluetooth devices to synchronize the time",
    )
    parser.add_argument(
        "-u",
        "--user",
        type=str,
        help="MQTT username",
    )
    parser.add_argument(
        "-ws",
        "--enable_websocket",
        type=int,
        help="Enable (1) or disable (0) WebSocket (default: 0)",
    )
    return parser.parse_args()


def read_configuration(config_path: Path) -> dict:
    """Read a Theengs Gateway configuration from a file.

    If the path doesn't exist, the function returns the default configuration.
    """
    try:
        with config_path.open(encoding="utf-8") as config_file:
            configuration = json.load(config_file)
    except OSError:
        configuration = DEFAULT_CONFIG
    except json.JSONDecodeError as exception:
        msg = f"Malformed configuration file {config_path.resolve()}: {exception}"
        raise SystemExit(msg) from exception

    return configuration


def write_configuration(configuration: dict, config_path: Path) -> None:
    """Write a Theengs Gateway configuration to a file."""
    try:
        with config_path.open(encoding="utf-8", mode="w") as config_file:
            config_file.write(
                json.dumps(configuration, sort_keys=True, indent=4),
            )
    except OSError as exception:
        msg = f"Unable to write configuration file {config_path.resolve()}"
        raise SystemExit(msg) from exception


def merge_args_with_config(config: dict, args: argparse.Namespace) -> None:
    """Merge command-line arguments into configuration.

    Command-line arguments override the corresponding configuration if they are set.
    Lists and dicts are merged.
    """
    for key, value in args.__dict__.items():
        if value is not None:
            if isinstance(value, list):
                if key in {"bindkeys", "identities"}:
                    config[key].update(
                        dict(zip(value[::2], value[1::2])),
                    )
                else:
                    config[key].extend(
                        element for element in value if element not in config[key]
                    )
            elif key != "config":
                config[key] = value



"""TheengsGateway - Decode things and devices and publish data to an MQTT broker.

Copyright: (c)Florian ROBERT

This file is part of TheengsGateway.

TheengsGateway is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TheengsGateway is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# mypy: disable-error-code="name-defined,attr-defined"
from __future__ import annotations

import asyncio
import json
import logging
import platform
import ssl
import struct
import sys
from contextlib import suppress
from datetime import datetime
from random import randrange
from threading import Thread
from time import time
from typing import TYPE_CHECKING, Dict, Union

from bleak import BleakError, BleakScanner
from bluetooth_clocks.exceptions import UnsupportedDeviceError
from bluetooth_clocks.scanners import find_clock
from bluetooth_numbers import company
from bluetooth_numbers.exceptions import No16BitIntegerError, UnknownCICError
from paho.mqtt import client as mqtt_client
from TheengsDecoder import decodeBLE

if TYPE_CHECKING:
    from pathlib import Path

    from bleak.backends.device import BLEDevice
    from bleak.backends.scanner import AdvertisementData

from .decryption import UnsupportedEncryptionError, create_decryptor
from .diagnose import diagnostics
from .privacy import resolve_private_address

if platform.system() == "Linux":
    from bleak.assigned_numbers import AdvertisementDataType
    from bleak.backends.bluezdbus.advertisement_monitor import OrPattern
    from bleak.backends.bluezdbus.scanner import BlueZScannerArgs

SECONDS_IN_DAY = 86400
ADVANCED_DATA = (
    "acts",
    "cidc",
    "cipher",
    "cont",
    "ctr",
    "encr",
    "manufacturerdata",
    "mic",
    "servicedata",
    "servicedatauuid",
)

LOG_LEVEL = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}

logger = logging.getLogger("BLEGateway")

DataJSONType = Dict[str, Union[str, int, float, bool]]


def get_address(data: DataJSONType) -> str:
    """Return the device address from a data JSON."""
    try:
        return data["mac"]  # type: ignore[return-value]
    except KeyError:
        return data["id"]  # type: ignore[return-value]


def add_manufacturer(
    data: DataJSONType,
) -> None:
    """Add the name of the manufacturer based on the company ID."""
    if "manufacturerdata" in data:
        company_id = int.from_bytes(
            bytes.fromhex(data["manufacturerdata"][:4]),  # type: ignore[index]
            byteorder="little",
        )
        with suppress(No16BitIntegerError, UnknownCICError):
            data["mfr"] = company[company_id]


class Gateway:
    """BLE to MQTT gateway class."""

    def __init__(
        self,
        configuration: dict,
    ) -> None:
        self.configuration = configuration
        self.stopped = False
        self.clock_updates: dict[str, float] = {}
        self.published_messages = 0
        self.discovered_trackers: dict[str, int] = {}

    def connect_mqtt(self) -> None:
        """Connect to MQTT broker."""

        def on_connect(
            client,  # noqa: ANN001
            userdata,  # noqa: ANN001,ARG001
            flags,  # noqa: ANN001,ARG001
            reason_code,  # noqa: ANN001
            properties,  # noqa: ANN001,ARG001
        ) -> None:
            if reason_code == 0:
                logger.info("Connected to MQTT broker")
                client.publish(
                    self.configuration["lwt_topic"],
                    "online",
                    qos=0,
                    retain=True,
                )
                self.subscribe(self.configuration["subscribe_topic"])
            else:
                logger.error(
                    "Failed to connect to MQTT broker %s:%d reason code: %d",
                    self.configuration["host"],
                    self.configuration["port"],
                    reason_code,
                )
                self.client.connect(
                    self.configuration["host"],
                    self.configuration["port"],
                )

        def on_disconnect(
            client,  # noqa: ANN001,ARG001
            userdata,  # noqa: ANN001,ARG001
            flags,  # noqa: ANN001,ARG001
            reason_code,  # noqa: ANN001
            properties,  # noqa: ANN001,ARG001
        ) -> None:
            if reason_code == 0:
                logger.info("Disconnected from MQTT broker")
            else:
                logger.error(
                    "Disconnected from MQTT broker with reason code = %d", reason_code
                )

        if self.configuration["enable_websocket"]:
            self.client = mqtt_client.Client(
                mqtt_client.CallbackAPIVersion.VERSION2, transport="websockets"
            )
        else:
            self.client = mqtt_client.Client(mqtt_client.CallbackAPIVersion.VERSION2)

        if self.configuration["enable_tls"]:
            self.client.tls_set(
                cert_reqs=ssl.CERT_REQUIRED,
                tls_version=ssl.PROTOCOL_TLS,
            )

        self.client.username_pw_set(
            self.configuration["user"],
            self.configuration["pass"],
        )
        self.client.will_set(
            self.configuration["lwt_topic"],
            "offline",
            qos=0,
            retain=True,
        )
        self.client.on_connect = on_connect
        self.client.on_disconnect = on_disconnect
        try:
            self.client.connect(
                self.configuration["host"],
                self.configuration["port"],
            )
        except Exception:
            logger.exception("Connection error")

    def disconnect_mqtt(self) -> None:
        """Disconnect from the MQTT broker."""
        self.client.publish(
            self.configuration["lwt_topic"],
            "offline",
            qos=0,
            retain=True,
        )
        self.client.disconnect()

    def subscribe(self, sub_topic: str) -> None:
        """Subscribe to MQTT topic <sub_topic>."""

        def on_message(client, userdata, msg) -> None:  # noqa: ANN001,ARG001
            logger.info(
                "Received `%s` from `%s` topic",
                msg.payload.decode(),
                msg.topic,
            )
            try:
                msg_json = json.loads(msg.payload.decode())
            except (json.JSONDecodeError, UnicodeDecodeError) as exception:
                logger.warning(
                    "Invalid JSON message %s: %s", msg.payload.decode(), exception
                )
                return

            try:
                msg_json["id"] = self.rpa2id(msg_json["id"])
            except KeyError:
                logger.warning(
                    "JSON message %s doesn't contain id", msg.payload.decode()
                )
                return

            self.decode_advertisement(msg_json)

        self.client.subscribe(sub_topic)
        self.client.on_message = on_message
        logger.info("Subscribed to %s", sub_topic)

    def hass_presence(self, decoded_json) -> None:  # noqa: ANN001
        """Add presence information to the decoded JSON."""
        rssi = decoded_json.get("rssi", 0)
        if not rssi:
            return
        txpower = decoded_json.get("txpower", 0)
        if txpower >= 0:
            txpower = -59  # if tx power is not found we set a default calibration value
        logger.debug("rssi: %d, txpower: %d", rssi, txpower)
        ratio = rssi / txpower
        if ratio < 1.0:  # noqa: PLR2004
            distance = pow(ratio, 10)
        else:
            distance = 0.89976 * pow(ratio, 7.7095) + 0.111
        decoded_json["distance"] = distance

    def publish(
        self,
        msg,  # noqa: ANN001
        pub_topic=None,  # noqa: ANN001
        retain=False,  # noqa: ANN001,FBT002
    ) -> None:
        """Publish <msg> to MQTT topic <pub_topic>."""
        if not pub_topic:
            pub_topic = self.configuration["publish_topic"]

        result = self.client.publish(pub_topic, msg, 0, retain)
        status = result[0]
        if status == 0:
            logger.debug("Sent `%s` to topic `%s`", msg, pub_topic)
            self.published_messages = self.published_messages + 1
        else:
            logger.error("Failed to send message to topic %s", pub_topic)

    def add_clock(self, address: str) -> None:
        """Register clock to synchronize its time later."""
        if (
            address in self.configuration["time_sync"]
            and address not in self.clock_updates
        ):
            # Add a random time in the last day as a starting point
            # for the daily update.
            # This prevents the gateway from connecting to all clocks
            # at the same time.
            start_time = time() - randrange(SECONDS_IN_DAY)  # noqa: S311
            self.clock_updates[address] = start_time
            logger.info(
                "Found device %s, synchronizing time daily beginning from %s",
                address,
                datetime.fromtimestamp(  # noqa: DTZ006
                    start_time + SECONDS_IN_DAY,
                ).strftime(
                    "%Y-%m-%d %H:%M:%S",
                ),
            )

    async def update_clock_times(self) -> None:
        """Update time for all registered clocks."""
        # Make a copy of the dictionary because we're changing it in the loop.
        for address, timestamp in self.clock_updates.copy().items():
            if time() - timestamp > SECONDS_IN_DAY:
                logger.info("Synchronizing time for clock %s...", address)

                # Find clock and try to synchronize the time
                try:
                    logger.info("Scanning for clock %s...", address)
                    clock = await find_clock(
                        address,
                        self.configuration["ble_scan_time"],
                    )
                    if clock:
                        logger.info(
                            "Writing time to %s device...",
                            clock.DEVICE_TYPE,
                        )
                        await clock.set_time(
                            ampm=bool(self.configuration["time_format"]),
                        )
                        logger.info("Synchronized time")
                    else:
                        logger.warning("Didn't find device %s.", address)
                except UnsupportedDeviceError:
                    logger.exception("Unsupported clock")
                    # There's no point in retrying for an unsupported device.
                    del self.clock_updates[address]
                    # Just continue with the next device.
                    continue
                except asyncio.exceptions.TimeoutError:
                    logger.exception("Can't connect to clock %s.", address)
                except BleakError:
                    logger.exception("Can't write to clock %s.", address)
                except AttributeError:
                    logger.exception(
                        "Can't get attribute from clock %s.",
                        address,
                    )

                # Register current time for this address
                this_time = time()
                self.clock_updates[address] = this_time
                logger.info(
                    "Synchronizing time with %s again on %s",
                    address,
                    datetime.fromtimestamp(  # noqa: DTZ006
                        this_time + SECONDS_IN_DAY,
                    ).strftime("%Y-%m-%d %H:%M:%S"),
                )

    def check_tracker_timeout(self) -> None:
        """Check if tracker timeout is over timeout limit."""
        # If the timestamp is later than current time minus tracker_timeout
        # Publish offline message
        for address, timestamp in self.discovered_trackers.copy().items():
            if (
                round(time()) - timestamp >= self.configuration["tracker_timeout"]
                and timestamp != 0
                and (
                    self.configuration["discovery"]
                    or self.configuration["general_presence"]
                )
            ):
                message = json.dumps({"id": address, "presence": "absent"})

                self.publish(
                    message,
                    self.configuration["publish_topic"]
                    + "/"
                    + address.replace(":", ""),
                )
                self.discovered_trackers[address] = 0

    async def ble_scan_loop(self) -> None:
        """Scan for BLE devices."""
        scanner_kwargs = {"scanning_mode": self.configuration["scanning_mode"]}

        if platform.system() == "Linux":
            if self.configuration["scanning_mode"] == "passive":
                # Passive scanning with BlueZ needs at least one or_pattern.
                # The following matches all devices.
                scanner_kwargs["bluez"] = BlueZScannerArgs(
                    or_patterns=[
                        OrPattern(0, AdvertisementDataType.FLAGS, b"\x06"),
                        OrPattern(0, AdvertisementDataType.FLAGS, b"\x1a"),
                    ],
                )  # type: ignore[assignment]
            elif self.configuration["scanning_mode"] == "active":
                # Disable duplicate detection of advertisement data.
                # Without this parameter non-compliant devices such as the
                # TP357/8/9 return multiple keys in manufacturer data and
                # we don't know which is the most recent data, so the sensor
                # values don't update.
                scanner_kwargs["bluez"] = BlueZScannerArgs(
                    filters={"DuplicateData": True},
                )  # type: ignore[assignment]

        if self.configuration["adapter"]:
            scanner_kwargs["adapter"] = self.configuration["adapter"]

        scanner_kwargs["detection_callback"] = self.detection_callback  # type: ignore[assignment]
        scanner = BleakScanner(**scanner_kwargs)  # type: ignore[arg-type]
        logger.info("Starting BLE scan")
        while not self.stopped:
            if self.client.is_connected():
                self.published_messages = 0
                try:
                    await scanner.start()
                    await asyncio.sleep(self.configuration["ble_scan_time"])
                    await scanner.stop()
                except BleakError as error:
                    logger.exception(error, exc_info=False)  # noqa: TRY401
                    sys.exit()
                    self.stopped = True
                logger.info(
                    "Sent %s messages to MQTT",
                    self.published_messages,
                )
                await asyncio.sleep(
                    self.configuration["ble_time_between_scans"],
                )

                # Update time for all clocks once a day
                await self.update_clock_times()
            else:
                await asyncio.sleep(5.0)

        logger.error("BLE scan loop stopped")

    def detection_callback(
        self,
        device: BLEDevice,
        advertisement_data: AdvertisementData,
    ) -> None:
        """Detect device in received advertisement data."""
        logger.debug("%s:%s", device.address, advertisement_data)

        # Try to resolve private addresses with known IRKs
        address = self.rpa2id(device.address)

        # Try to add the device to dictionary of clocks to synchronize time.
        self.add_clock(address)

        data_json: DataJSONType = {}

        if advertisement_data.manufacturer_data:
            # Only look at the first manufacturer data in the advertisement
            company_id = list(  # noqa: RUF015
                advertisement_data.manufacturer_data.keys(),
            )[0]
            manufacturer_data = list(  # noqa: RUF015
                advertisement_data.manufacturer_data.values(),
            )[0]
            dstr = str(struct.pack("<H", company_id).hex())
            dstr += str(manufacturer_data.hex())
            data_json["manufacturerdata"] = dstr

        if advertisement_data.local_name:
            data_json["name"] = advertisement_data.local_name

        if advertisement_data.service_data:
            data_json["id"] = address
            data_json["rssi"] = advertisement_data.rssi
            # Try to decode advertisement with service data for each UUID separately.
            for uuid, data in advertisement_data.service_data.items():
                # Copy data JSON because it gets manipulated while decoding
                data_json_copy = data_json.copy()
                data_json_copy["servicedatauuid"] = uuid[4:8]
                data_json_copy["servicedata"] = data.hex()
                self.decode_advertisement(data_json_copy)
        elif data_json:
            data_json["id"] = address
            data_json["rssi"] = advertisement_data.rssi
            self.decode_advertisement(data_json)

    def decode_advertisement(
        self,
        data_json: DataJSONType,
    ) -> None:
        """Decode device from data JSON."""
        decoded_json = decodeBLE(json.dumps(data_json))

        if decoded_json:
            decoded_json = json.loads(decoded_json)

            decoded_json = self.process_prmacs(
                decoded_json,
            )

            # Only process if the device is not a random mac address
            if decoded_json["type"] != "RMAC":
                # Only add manufacturer if device is compliant and no beacon
                if (
                    decoded_json.get("cidc", True)
                    and decoded_json.get("type", "UNIQ") != "BCON"
                ):
                    add_manufacturer(data_json)

                if self.configuration["presence"]:
                    self.hass_presence(decoded_json)

                # Handle encrypted payload
                if decoded_json.get("encr", 0) > 0:
                    decoded_json = self.handle_encrypted_advertisement(
                        data_json,
                        decoded_json,
                    )

                # Remove advanced data
                if not self.configuration["publish_advdata"]:
                    for key in ADVANCED_DATA:
                        decoded_json.pop(key, None)

                if self.configuration["discovery"]:
                    self.publish_device_info(
                        decoded_json,
                    )  # Publish sensor data to Home Assistant MQTT discovery
                else:
                    self.publish_json(decoded_json, decoded=True)

                # Check tracker timeouts
                self.check_tracker_timeout()

        elif self.configuration["publish_all"]:
            add_manufacturer(data_json)
            self.publish_json(data_json, decoded=False)

    def rpa2id(self, address: str) -> str:
        """Replace a random private address by its corresponding identity address.

        If there's no identity known for the address, this function just returns
        the original address.
        """
        for identity, irk in self.configuration["identities"].items():
            if resolve_private_address(address, irk):
                logger.debug(
                    "Using identity address %s instead of random private address %s",
                    identity,
                    address,
                )
                return identity

        return address

    def process_prmacs(
        self,
        decoded_json: DataJSONType,
    ) -> DataJSONType:
        """Process potential RMACs."""
        if "prmac" in decoded_json:
            address = decoded_json.get("id")
            if address in self.configuration["identities"]:
                decoded_json.pop("prmac", None)
            else:
                decoded_json["type"] = "RMAC"

        return decoded_json

    def publish_json(
        self,
        data_json: DataJSONType,
        decoded: bool,  # noqa: FBT001
    ) -> None:
        """Publish JSON data to MQTT."""
        # publish general presence "present" if tracker and general_presence true
        if "track" in data_json:
            if (
                data_json["id"] not in self.discovered_trackers
                or (
                    data_json["id"] in self.discovered_trackers
                    and self.discovered_trackers[str(data_json["id"])] == 0
                )
            ) and self.configuration["general_presence"]:
                message = json.dumps({"id": data_json["id"], "presence": "present"})
                self.publish(
                    message,
                    self.configuration["publish_topic"]
                    + "/"
                    + get_address(data_json).replace(":", ""),
                )
            self.discovered_trackers[str(data_json["id"])] = round(time())
            logger.debug("Discovered Trackers: %s", self.discovered_trackers)

        # Remove "track" if PUBLISH_ADVDATA is 0
        if not self.configuration["publish_advdata"] and "track" in data_json:
            data_json.pop("track", None)

        message = json.dumps(data_json)
        self.publish(
            message,
            self.configuration["publish_topic"]
            + "/"
            + get_address(data_json).replace(":", ""),
        )
        if decoded and self.configuration["presence"]:
            self.publish(
                message,
                self.configuration["presence_topic"],
            )

    def handle_encrypted_advertisement(
        self,
        data_json: DataJSONType,
        decoded_json: DataJSONType,
    ) -> DataJSONType:
        """Handle encrypted advertisement."""
        try:
            bindkey = bytes.fromhex(
                self.configuration["bindkeys"][get_address(decoded_json)],
            )
            decryptor = create_decryptor(
                decoded_json["encr"],  # type: ignore[arg-type]
            )
            decrypted_data = decryptor.decrypt(
                bindkey,
                get_address(decoded_json),
                decoded_json,
            )
            decryptor.replace_encrypted_data(
                decrypted_data,
                data_json,
                decoded_json,
            )

            # Keep encrypted properties
            cipher = decoded_json["cipher"]
            mic = decoded_json["mic"]
            ctr = decoded_json["ctr"]

            # Re-decode advertisement, this time unencrypted
            decoded_json = decodeBLE(json.dumps(data_json))
            if decoded_json:
                decoded_json = json.loads(decoded_json)  # type: ignore[arg-type]
            else:
                logger.exception(
                    "Decrypted payload not supported: `%s`",
                    data_json["servicedata"],
                )

            # Re-add encrypted properties
            decoded_json["cipher"] = cipher
            decoded_json["mic"] = mic
            decoded_json["ctr"] = ctr
        except KeyError:
            logger.exception(
                "Can't find bindkey for %s.", get_address(decoded_json), exc_info=False
            )
        except UnsupportedEncryptionError:
            logger.exception(
                "Unsupported encrypted device %s of model %s",
                get_address(decoded_json),
                decoded_json["model_id"],
            )
        except ValueError:
            logger.exception("Decryption failed")
        finally:
            return decoded_json  # noqa: B012


def run(configuration: dict, config_path: Path) -> None:
    """Run BLE gateway."""
    if configuration["discovery"]:
        from .discovery import DiscoveryGateway

        gw = DiscoveryGateway(configuration)
    else:
        gw = Gateway(configuration)  # type: ignore[assignment]

    logging.basicConfig()
    log_level = LOG_LEVEL[configuration["log_level"].upper()]
    logger.setLevel(log_level)

    loop = asyncio.get_event_loop()

    if log_level == logging.DEBUG:
        asyncio.run(diagnostics(config_path))
    thread = Thread(target=loop.run_forever, daemon=True)
    thread.start()
    if configuration["ble"]:
        asyncio.run_coroutine_threadsafe(gw.ble_scan_loop(), loop)

    gw.connect_mqtt()

    try:
        gw.client.loop_forever()
    except (KeyboardInterrupt, SystemExit):
        gw.disconnect_mqtt()
        gw.stopped = True
        loop.call_soon_threadsafe(loop.stop)
        thread.join()



"""Setup script for TheengsGateway package."""
from pathlib import Path

from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name="TheengsGateway",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Theengs",
    url="https://github.com/theengs/gateway",
    license="GPL-3.0 License",
    package_dir={"TheengsGateway": "TheengsGateway"},
    packages=["TheengsGateway"],
    scripts=["bin/TheengsGateway"],
    include_package_data=True,
    install_requires=[
        "bleak>=0.19.0",
        'bluetooth-adapters>=0.15.3; python_version>="3.9"',
        "bluetooth-clocks<1.0",
        "bluetooth-numbers>=1.0,<2.0",
        "importlib-metadata",
        "paho-mqtt>=2.0.0,<3.0.0",
        "pycryptodomex>=3.18.0",
        "TheengsDecoder>=1.7.0",
    ],
    use_scm_version={"version_scheme": "no-guess-dev"},
    setup_requires=["setuptools_scm"],
)



"""Privacy module for Theengs Gateway.

This module is used to resolve resolvable private addresses (RPAs) if you know their
corresponding identity resolving keys (IRKs).
"""
from __future__ import annotations

from base64 import b64decode

from Cryptodome.Cipher import AES

_MSB_MASK = 0b11000000
_MSB_RPA = 0b01000000


def resolve_private_address(address: str, irk: str) -> bool:
    """Return `True` if the address can be resolved with the IRK."""
    rpa = bytes.fromhex(address.replace(":", ""))
    # Format of a resolvable private address:
    # MSB                                          LSB
    # 01 Random part of prand      Hash value
    # <-   prand (24 bits)  -><-   hash (24 bits)   ->
    if rpa[0] & _MSB_MASK != _MSB_RPA:
        # Address is not an RPA
        return False

    prand = rpa[:3]
    hash_value = rpa[3:]
    try:
        # Suppose the key is in hex format
        key = bytes.fromhex(irk)[::-1]
    except ValueError:
        # If that doesn't work, try to decode it as Base64 format
        key = b64decode(irk)[::-1]
    cipher = AES.new(key, AES.MODE_ECB)
    localhash = cipher.encrypt(b"\x00" * 13 + prand)

    if localhash[13:] != hash_value:
        # 24 bits of local hash don't match the RPA's hash
        return False

    return True
